from ._custom_observables import CustomObservables
from ._identity import Identity
from ._polynomial import Polynomial
from ._time_delay import TimeDelay

__all__ = ["CustomObservables", "Identity", "Polynomial", "TimeDelay"]



from warnings import warn

from numpy import identity
from pydmd import DMDBase
from sklearn.utils.validation import check_is_fitted

from ._base import BaseRegressor


class DMDRegressor(BaseRegressor):
    """
    Wrapper for PyDMD regressors.

    Parameters
    ----------
    DMDRegressor: DMDBase subclass
        Regressor from PyDMD. Must extend the DMDBase class.
    """

    def __init__(self, regressor):
        if not isinstance(regressor, DMDBase):
            raise ValueError("regressor must be a subclass of DMDBase from pydmd.")
        super(DMDRegressor, self).__init__(regressor)

    def fit(self, x, y=None):
        """
        Parameters
        ----------
        x: numpy ndarray, shape (n_examples, n_features)
            Measurement data to be fit.

        Returns
        -------
        self: returns a fit ``DMDRegressor`` instance
        """
        if y is not None:
            warn("pydmd regressors do not require the y argument when fitting.")
        # We transpose x because PyDMD assumes examples are columns, not rows
        self.regressor.fit(x.T)
        self.coef_ = self.regressor.predict(identity(x.shape[1])).T

        return self

    def predict(self, x):
        """
        Parameters
        ----------
        x: numpy ndarray, shape (n_examples, n_features)
            Measurement data upon which to base prediction.

        Returns
        -------
        y: numpy ndarray, shape (n_examples, n_features)
            Prediction of x one timestep in the future.

        """
        check_is_fitted(self, "coef_")
        return self.regressor.predict(x.T).T



import importlib
import pathlib

author = "Eurika Kaiser and Brian de Silva"
project = "pykoopman"  # package name


# no need to edit below this line

copyright = f"2020, {author}"

module = importlib.import_module(project)
version = release = getattr(module, "__version__")
# version = "0.0.1"

# The master toctree document.
master_doc = "index"

extensions = [
    "sphinxcontrib.apidoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx_nbexamples",
    "sphinx.ext.intersphinx",
]

apidoc_module_dir = f"../{project}"
apidoc_excluded_paths = ["tests"]
apidoc_toc_file = False

autodoc_default_options = {"members": True}
autodoc_member_order = "bysource"
autoclass_content = "init"

language = None

here = pathlib.Path(__file__).parent

if (here / "static/custom.css").exists():

    html_static_path = ["static"]

    def setup(app):
        app.add_stylesheet("custom.css")


exclude_patterns = ["build", "_build", "Thumbs.db", ".DS_Store"]
# pygments_style = "sphinx"

add_module_names = True
add_function_parentheses = False

html_theme = "sphinx_rtd_theme"
html_show_sourcelink = False
html_show_sphinx = False
html_show_copyright = True

default_role = "any"
html_sourcelink_suffix = ""

example_gallery_config = dict(
    dont_preprocess=True,
    examples_dirs=["../examples"],
    gallery_dirs=["examples"],
    pattern=".+.ipynb",
)

intersphinx_mapping = {
    "derivative": ("https://derivative.readthedocs.io/en/latest/", None)
}

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "pykoopman", "pykoopman Documentation", [author], 1)]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Extensions to the  Napoleon GoogleDocstring class ---------------------
# michaelgoerz.net/notes/extending-sphinx-napoleon-docstring-sections.html
from sphinx.ext.napoleon.docstring import GoogleDocstring  # noqa: E402


def parse_keys_section(self, section):
    return self._format_fields("Keys", self._consume_fields())


GoogleDocstring._parse_keys_section = parse_keys_section


def parse_attributes_section(self, section):
    return self._format_fields("Attributes", self._consume_fields())


GoogleDocstring._parse_attributes_section = parse_attributes_section


def parse_class_attributes_section(self, section):
    return self._format_fields("Class Attributes", self._consume_fields())


GoogleDocstring._parse_class_attributes_section = parse_class_attributes_section


def patched_parse(self):
    """
    we now patch the parse method to guarantee that the the above methods are
    assigned to the _section dict
    """
    self._sections["keys"] = self._parse_keys_section
    self._sections["class attributes"] = self._parse_class_attributes_section
    self._unpatched_parse()


GoogleDocstring._unpatched_parse = GoogleDocstring._parse
GoogleDocstring._parse = patched_parse



import numpy as np
from sklearn.utils import check_array as skl_check_array

T_DEFAULT = object()


def validate_input(x, t=T_DEFAULT):
    if not isinstance(x, np.ndarray):
        raise ValueError("x must be array-like")
    elif x.ndim == 1:
        x = x.reshape(-1, 1)
    x = check_array(x)

    if t is not T_DEFAULT:
        if t is None:
            raise ValueError("t must be a scalar or array-like.")
        # Apply this check if t is a scalar
        elif np.ndim(t) == 0 and (isinstance(t, int) or isinstance(t, float)):
            if t <= 0:
                raise ValueError("t must be positive")
        # Only apply these tests if t is array-like
        elif isinstance(t, np.ndarray):
            if not len(t) == x.shape[0]:
                raise ValueError("Length of t should match x.shape[0].")
            if not np.all(t[:-1] < t[1:]):
                raise ValueError("Values in t should be in strictly increasing order.")
        else:
            raise ValueError("t must be a scalar or array-like.")

    return x


def check_array(x, **kwargs):
    if np.iscomplexobj(x):
        return skl_check_array(x.real, **kwargs) + 1j * skl_check_array(
            x.imag, **kwargs
        )
    else:
        return skl_check_array(x, **kwargs)


def drop_nan_rows(arr, *args):
    """
    Remove rows in all inputs for which `arr` has `_np.nan` entries.

    Parameters
    ----------
    arr : numpy.ndarray
        Array whose rows are checked for nan entries.
        Any rows containing nans are removed from ``arr`` and all arguments
        passed via ``args``.
    *args : variable length argument list of numpy.ndarray
        Additional arrays from which to remove rows.
        Each argument should have the same number of rows as ``arr``.

    Returns
    -------
    arrays : tuple of numpy.ndarray
        Arrays with nan rows dropped.
        The first entry corresponds to ``arr`` and all following entries
        to ``*args``.
    """
    nan_inds = np.isnan(arr).any(axis=1)
    return (arr[~nan_inds], *[arg[~nan_inds] for arg in args])



import pytest
from numpy import hstack
from numpy import iscomplexobj
from numpy import linspace
from numpy import stack
from numpy.testing import assert_allclose
from numpy.testing import assert_array_equal
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from pykoopman.observables import CustomObservables
from pykoopman.observables import Identity
from pykoopman.observables import Polynomial
from pykoopman.observables import TimeDelay


@pytest.fixture
def data_small():
    t = linspace(0, 5, 10)
    return stack((t, t ** 2), axis=1)


@pytest.mark.parametrize(
    "observables",
    [
        Identity(),
        Polynomial(),
        TimeDelay(),
        pytest.lazy_fixture("data_custom_observables"),
    ],
)
def test_if_fitted(observables, data_random):
    x = data_random
    with pytest.raises(NotFittedError):
        observables.transform(x)

    with pytest.raises(NotFittedError):
        observables.inverse(x)

    with pytest.raises(NotFittedError):
        observables.get_feature_names()

    observables.fit(x)
    check_is_fitted(observables)


@pytest.mark.parametrize(
    "observables",
    [
        Identity(),
        Polynomial(),
        Polynomial(degree=1),
        Polynomial(degree=4),
        Polynomial(include_bias=False),
        Polynomial(degree=3, include_bias=False),
        pytest.lazy_fixture("data_custom_observables"),
    ],
)
def test_inverse(observables, data_random):
    x = data_random
    assert_allclose(observables.inverse(observables.fit_transform(x)), x)


def test_time_delay_inverse(data_random):
    x = data_random
    delay = 2
    n_delays = 3
    n_deleted_rows = delay * n_delays

    observables = TimeDelay(delay=delay, n_delays=n_delays)
    y = observables.fit_transform(x)
    # First few rows of x are deleted which don't have enough
    # time history
    assert_array_equal(observables.inverse(y), x[n_deleted_rows:])


def test_bad_polynomial_inputs():
    with pytest.raises(ValueError):
        Polynomial(degree=0)


def test_bad_custom_observables_inputs():
    # One too few names
    observables = [lambda x: x, lambda x: x ** 2, lambda x: 0 * x, lambda x, y: x * y]
    observable_names = [lambda s: f"{s}^2", lambda: str(0), lambda s, t: f"{s} {t}"]

    with pytest.raises(ValueError):
        CustomObservables(observables, observable_names=observable_names)


def test_custom_observables_transform(data_small):
    x = data_small

    observables = [lambda x: x ** 2]
    y = CustomObservables(observables).fit_transform(x)

    # Identity is automatically prepended to custom observables
    assert_array_equal(y, hstack((x, x ** 2)))


@pytest.mark.parametrize("delay, n_delays", [(3, 2), (1, 5)])
def test_time_delay_output_shape(data_random, delay, n_delays):
    x = data_random
    y = TimeDelay(delay=delay, n_delays=n_delays).fit_transform(x)

    assert y.shape == (x.shape[0] - delay * n_delays, (n_delays + 1) * x.shape[1])


def test_time_delay_transform_matches_input(data_random):
    x = data_random

    observables = TimeDelay(delay=2, n_delays=4)
    observables.fit(x)

    y = observables.transform(x)
    assert_array_equal(y[1], x[[9, 7, 5, 3, 1]].flatten())


@pytest.mark.parametrize(
    "observables, expected_default_names, expected_custom_names",
    [
        (Identity(), ["x0", "x1"], ["x", "y"]),
        (
            Polynomial(degree=2),
            ["1", "x0", "x1", "x0^2", "x0 x1", "x1^2"],
            ["1", "x", "y", "x^2", "x y", "y^2"],
        ),
        (
            TimeDelay(delay=2, n_delays=2),
            ["x0(t)", "x1(t)", "x0(t-2dt)", "x1(t-2dt)", "x0(t-4dt)", "x1(t-4dt)"],
            ["x(t)", "y(t)", "x(t-2dt)", "y(t-2dt)", "x(t-4dt)", "y(t-4dt)"],
        ),
        (
            pytest.lazy_fixture("data_custom_observables"),
            ["x0", "x1", "x0", "x1", "x0^2", "x1^2", "0", "0", "x0 x1"],
            ["x", "y", "x", "y", "x^2", "y^2", "0", "0", "x y"],
        ),
    ],
)
def test_feature_names(
    observables, expected_default_names, expected_custom_names, data_small
):
    x = data_small

    observables.fit(x)
    assert observables.get_feature_names() == expected_default_names

    custom_names = ["x", "y"]
    assert (
        observables.get_feature_names(input_features=custom_names)
        == expected_custom_names
    )


@pytest.mark.parametrize(
    "observables",
    [
        Identity(),
        Polynomial(),  # TODO: rewrite Polynomial to handle cplx. nums.
        TimeDelay(),
        pytest.lazy_fixture("data_custom_observables"),
    ],
)
def test_complex_data(data_random_complex, observables):
    x = data_random_complex
    y = observables.fit_transform(x)

    assert iscomplexobj(y)



"""Tests for differentiation methods."""
import numpy as np
import pytest
from derivative import dxdt

from pykoopman.differentiation import Derivative


@pytest.fixture
def data_1D_quadratic():
    t = np.linspace(0, 5, 100)
    x = t.reshape(-1, 1) ** 2
    x_dot = 2 * t.reshape(-1, 1)

    return x, t, x_dot


@pytest.fixture
def data_1D_bad_shape():
    t = np.linspace(0, 5, 100)
    x = t ** 2
    x_dot = 2 * t

    return x, t, x_dot


@pytest.fixture
def data_2D_quadratic():
    t = np.linspace(0, 5, 100)
    x = np.zeros((len(t), 2))
    x[:, 0] = t ** 2
    x[:, 1] = -(t ** 2)

    x_dot = np.zeros_like(x)
    x_dot[:, 0] = 2 * t
    x_dot[:, 1] = -2 * t

    return x, t, x_dot


@pytest.fixture(params=["data_1D_quadratic", "data_1D_bad_shape", "data_2D_quadratic"])
def data(request):
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    "kws",
    [
        dict(kind="spectral"),
        dict(kind="spline", s=1e-2),
        dict(kind="trend_filtered", order=0, alpha=1e-2),
        dict(kind="finite_difference", k=1),
        dict(kind="savitzky_golay", order=3, left=1, right=1),
    ],
)
def test_derivative_package_equivalence(data, kws):
    x, t, _ = data

    x_dot_pykoopman = Derivative(**kws)(x, t)
    x_dot_derivative = dxdt(x, t, axis=0, **kws).reshape(x_dot_pykoopman.shape)

    np.testing.assert_array_equal(x_dot_pykoopman, x_dot_derivative)


def test_bad_t_values(data_1D_quadratic):
    x, t, _ = data_1D_quadratic

    method = Derivative(kind="finite_difference", k=1)

    with pytest.raises(ValueError):
        method(x, t=-1)

    with pytest.raises(ValueError):
        method(x, t[:5])

    with pytest.raises(ValueError):
        inds = np.arange(len(t))
        # Swap two time entries
        inds[[0, 1]] = inds[[1, 0]]
        method(x, t[inds])


def test_accuracy(data):
    x, t, x_dot = data

    method = Derivative(kind="finite_difference", k=1)
    x_dot_method = method(x, t)

    if x_dot.ndim == 1:
        x_dot = x_dot.reshape(-1, 1)

    # Ignore endpoints
    np.testing.assert_allclose(x_dot[1:-1], x_dot_method[1:-1])



"""
Base class for specific observable classes
"""
import abc

from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted


class BaseObservables(TransformerMixin, BaseEstimator):
    """
    Base class for observable classes.

    Forces subclasses to implement 'fit', 'transform', and 'get_feature_names' functions
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def fit(self, X, y=None):
        """
        Does not do anything for now.

        Parameters
        ----------
        X : numpy ndarray with shape [n_samples, n_features]

        Returns
        -------
        self : instance
        """
        raise NotImplementedError

    @abc.abstractmethod
    def transform(self, X):
        """
        Transforms data.

        Parameters
        ----------
        X : numpy ndarray with shape [n_samples, n_features]
            Data X is transformed row-wise.

        Returns
        -------
        XT : numpy ndarray with shape [n_samples, n_output_features]
             Transformed data X into observables XT, where the number of
             observables or features is n_output_features.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_feature_names(self, input_features=None):
        """
        Return names of observables.

        Parameters
        ----------
        input_features : list of string of length n_features, optional
            Default list is "x0", "x1", ..., "xn", where n = n_features.

        Returns
        -------
        output_feature_names : list of string of length n_output_features
        """
        raise NotImplementedError

    @abc.abstractmethod
    def inverse(self, y):
        """
        Compute inverse mapping satisfying
        :code:`self.inverse(self.transform(x)) == x`

        Parameters
        ----------
        y: array-like, shape (n_samples, n_output_features)
            Data to which to apply the inverse.
            Must have the same number of features as the transformed data
        """
        raise NotImplementedError

    @property
    def size(self):
        check_is_fitted(self)
        return self.n_output_features_



"""
Wrapper classes for differentiation methods from the :doc:`derivative:index` package.

Some default values used here may differ from those used in :doc:`derivative:index`.
"""
from derivative import dxdt
from numpy import arange
from sklearn.base import BaseEstimator

from ..common import validate_input


class Derivative(BaseEstimator):
    """
    Wrapper class for differentiation classes from the :doc:`derivative:index` package.
    This class is meant to provide all the same functionality as the
    `dxdt <https://derivative.readthedocs.io/en/latest/api.html\
        #derivative.differentiation.dxdt>`_ method.

    This class includes a :meth:`__call__` method.

    Parameters
    ----------
    derivative_kws: dictionary, optional
        Keyword arguments to be passed to the
        `dxdt <https://derivative.readthedocs.io/en/latest/api.html\
        #derivative.differentiation.dxdt>`_
        method.

    Notes
    -----
    See the `derivative documentation <https://derivative.readthedocs.io/en/latest/>`_
    for acceptable keywords.
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def set_params(self, **params):
        """
        Set the parameters of this estimator.
        Modification of the pysindy method to allow unknown kwargs. This allows using
        the full range of derivative parameters that are not defined as member variables
        in sklearn grid search.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        else:
            self.kwargs.update(params)

        return self

    def get_params(self, deep=True):
        """Get parameters."""
        params = super().get_params(deep)

        if isinstance(self.kwargs, dict):
            params.update(self.kwargs)

        return params

    def __call__(self, x, t):
        """
        Perform numerical differentiation by calling the ``dxdt`` method.

        Paramters
        ---------
        x: np.ndarray, shape (n_samples, n_features)
            Data to be differentiated. Rows should correspond to different
            points in time and columns to different features.

        t: np.ndarray, shape (n_samples, )
            Time points for each sample (row) in ``x``.

        Returns
        -------
        x_dot: np.ndarray, shape (n_samples, n_features)
        """
        x = validate_input(x, t=t)

        if isinstance(t, (int, float)):
            if t < 0:
                raise ValueError("t must be a positive constant or an array")
            t = arange(x.shape[0]) * t

        return dxdt(x, t, axis=0, **self.kwargs)



import numpy as np
from sklearn.base import BaseEstimator


class FiniteDifference(BaseEstimator):
    def __init__(self, order=1):
        self.order = order

    def __call__(self, x, t=1):
        return np.gradient(x)



"""
Linear observables
"""
from sklearn.utils.validation import check_is_fitted

from ._base import BaseObservables


class Identity(BaseObservables):
    """
    A dummy observables class that simply returns its input.
    """

    def __init__(self):
        super().__init__()

    def fit(self, x, y=None):
        """
        Fit to measurement data.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Measurement data to be fit.

        y: None
            Dummy parameter retained for sklearn compatibility.

        Returns
        -------
        self: a fit :class:`pykoopman.observables.Identity` instance
        """
        # TODO: validate input
        self.n_input_features_ = self.n_output_features_ = x.shape[1]

        return self

    def transform(self, x):
        """
        Apply Identity transformation to data.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Measurement data to be transformed.

        Returns
        -------
        y: array-like, shape (n_samples, n_input_features)
            Transformed data (same as x in this case).
        """
        # TODO validate input
        check_is_fitted(self, "n_input_features_")
        return x

    def inverse(self, y):
        """
        Invert the transformation.

        This function satisfies
        :code:`self.inverse(self.transform(x)) == x`

        Parameters
        ----------
        y: array-like, shape (n_samples, n_output_features)
            Data to which to apply the inverse.
            Must have the same number of features as the transformed data

        Returns
        -------
        x: array-like, shape (n_samples, n_input_features)
            Output of inverse map applied to y.
            In this case, x is identical to y.
        """
        # TODO: validate input
        check_is_fitted(self, "n_input_features_")
        return y

    def get_feature_names(self, input_features=None):
        """
        Get the names of the output features.

        Parameters
        ----------
        input_features: list of string, length n_input_features,\
         optional (default None)
            String names for input features, if available. By default,
            the names "x0", "x1", ... ,"xn_input_features" are used.

        Returns
        -------
        output_feature_names: list of string, length n_ouput_features
            Output feature names.
        """
        check_is_fitted(self, "n_input_features_")
        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )
        return input_features



from sklearn.base import BaseEstimator


class BaseRegressor(BaseEstimator):
    """
    Wrapper class for PyKoopman regressors.

    Parameters
    ----------
    regressor: regressor object
        A regressor object implementing ``fit`` and ``predict`` methods.
    """

    def __init__(self, regressor):
        if not hasattr(regressor, "fit") or not callable(getattr(regressor, "fit")):
            raise AttributeError("regressor does not have a callable fit method")
        if not hasattr(regressor, "predict") or not callable(
            getattr(regressor, "predict")
        ):
            raise AttributeError("regressor does not have a callable predict method")

        self.regressor = regressor

    def fit(self, x, y):
        raise NotImplementedError

    def predict(self, x):
        raise NotImplementedError



from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass

from .koopman import Koopman
from .koopman_continuous import KoopmanContinuous


__all__ = [
    "Koopman",
    "KoopmanContinuous",
    "common",
    "differentiation",
    "observables",
    "regression",
]



"""Tests for pykoopman.regression objects and methods."""
import pytest

from pykoopman.regression import BaseRegressor


class RegressorWithoutFit:
    def __init__(self):
        pass

    def predict(self, x):
        return x


class RegressorWithoutPredict:
    def __init__(self):
        pass

    def fit(self, x):
        return self


@pytest.mark.parametrize(
    "regressor", [RegressorWithoutFit(), RegressorWithoutPredict()]
)
def test_bad_regressor_input(regressor):
    with pytest.raises(AttributeError):
        BaseRegressor(regressor)



from warnings import catch_warnings
from warnings import filterwarnings
from warnings import warn

from numpy import empty
from numpy import vstack
from pydmd import DMD
from pydmd import DMDBase
from sklearn.base import BaseEstimator
from sklearn.metrics import r2_score
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from .common import validate_input
from .observables import Identity
from .observables import TimeDelay
from .regression import BaseRegressor
from .regression import DMDRegressor


class Koopman(BaseEstimator):
    """
    Discrete-Time Koopman class.

    Parameters
    ----------
    observables: observables object, optional \
            (default :class:`pykoopman.observables.Identity`)
        Map(s) to apply to raw measurement data before estimating the
        Koopman operator.
        Must extend :class:`pykoopman.observables.BaseObservables`.
        The default option, :class:`pykoopman.observables.Identity` leaves
        the input untouched.

    regressor: regressor object, optional (default ``DMD``)
        The regressor used to learn the Koopman operator from the observables.
        ``regressor`` can either extend the
        :class:`pykoopman.regression.BaseRegressor`, or ``pydmd.DMDBase``.
        In the latter case, the pydmd object must have both a ``fit``
        and a ``predict`` method.

    quiet: booolean, optional (default False)
        Whether or not warnings should be silenced during fitting.

    Attributes
    ----------
    model: sklearn.pipeline.Pipeline
        Internal representation of the forward model.
        Applies the observables and the regressor.

    n_input_features_: int
        Number of input features before computing observables.

    n_output_features_: int
        Number of output features after computing observables.
    """

    def __init__(self, observables=None, regressor=None, quiet=False):
        if observables is None:
            observables = Identity()
        if regressor is None:
            regressor = DMD(svd_rank=2)
        if isinstance(regressor, DMDBase):
            regressor = DMDRegressor(regressor)
        elif not isinstance(regressor, (BaseRegressor)):
            raise TypeError("Regressor must be from valid class")

        self.observables = observables
        self.regressor = regressor
        self.quiet = quiet

    def fit(self, x):
        """
        Fit the Koopman model by learning an approximate Koopman operator.

        Parameters
        ----------
        x: numpy.ndarray, shape (n_samples, n_features)
            Measurement data to be fit. Each row should correspond to an example
            and each column a feature. It is assumed that examples are
            equi-spaced in time (i.e. a uniform timestep is assumed).

        Returns
        -------
        self: returns a fit ``Koopman`` instance
        """
        x = validate_input(x)

        steps = [
            ("observables", self.observables),
            ("regressor", self.regressor),
        ]
        self.model = Pipeline(steps)

        action = "ignore" if self.quiet else "default"
        with catch_warnings():
            filterwarnings(action, category=UserWarning)
            self.model.fit(x)

        self.n_input_features_ = self.model.steps[0][1].n_input_features_
        self.n_output_features_ = self.model.steps[0][1].n_output_features_

        return self

    def predict(self, x):
        """
        Predict the state one timestep in the future.

        Parameters
        ----------
        x: numpy.ndarray, shape (n_samples, n_input_features)
            Current state.

        Returns
        -------
        y: numpy.ndarray, shape (n_samples, n_input_features)
            Predicted state one timestep in the future.
        """
        check_is_fitted(self, "n_output_features_")
        return self.observables.inverse(self._step(x))

    def simulate(self, x0, n_steps=1):
        """
        Simulate an initial state forward in time with the learned Koopman
        model.

        Parameters
        ----------
        x0: numpy.ndarray, shape (n_input_features,) or \
                (n_consumed_samples + 1, n_input_features)
            Initial state from which to simulate.
            If using :code:`TimeDelay` observables, ``x0`` should contain
            enough examples to compute all required time delays,
            i.e. ``n_consumed_samples + 1``.

        n_steps: int, optional (default 1)
            Number of forward steps to be simulated.

        Returns
        -------
        y: numpy.ndarray, shape (n_steps, n_input_features)
            Simulated states.
            Note that ``y[0, :]`` is one timestep ahead of ``x0``.
        """
        check_is_fitted(self, "n_output_features_")
        # Could have an option to only return the end state and not all
        # intermediate states to save memory.
        y = empty((n_steps, self.n_input_features_), dtype=self.koopman_matrix.dtype)
        y[0] = self.predict(x0)

        if isinstance(self.observables, TimeDelay):
            n_consumed_samples = self.observables.n_consumed_samples
            for k in range(n_consumed_samples):
                y[k + 1] = self.predict(vstack((x0[k + 1 :], y[: k + 1])))

            for k in range(n_consumed_samples, n_steps - 1):
                y[k + 1] = self.predict(y[k - n_consumed_samples : k + 1])
        else:
            for k in range(n_steps - 1):
                y[k + 1] = self.predict(y[k])

        return y

    def score(self, x, y=None, cast_as_real=True, metric=r2_score, **metric_kws):
        """
        Score the model prediction for the next timestep.

        Parameters
        ----------
        x: numpy.ndarray, shape (n_samples, n_input_features)
            State measurements.
            Each row should correspond to the system state at some point
            in time.
            If ``y`` is not passed, then it is assumed that the examples are
            equi-spaced in time and are given in sequential order.
            If ``y`` is passed, then this assumption need not hold.

        y: numpy.ndarray, shape (n_samples, n_input_features), optional \
                (default None)
            State measurements one timestep in the future.
            Each row of this array should give the corresponding row in x advanced
            forward in time by one timestep.
            If None, the rows of ``x`` are used to construct ``y``.

        cast_as_real: bool, optional (default True)
            Whether to take the real part of predictions when computing the score.
            Many Scikit-learn metrics do not support complex numbers.

        metric: callable, optional (default ``r2_score``)
            The metric function used to score the model predictions.

        metric_kws: dict, optional
            Optional parameters to pass to the metric function.

        Returns
        -------
        score: float
            Metric function value for the model predictions at the next timestep.
        """
        check_is_fitted(self, "n_output_features_")
        x = validate_input(x)

        if isinstance(self.observables, TimeDelay):
            n_consumed_samples = self.observables.n_consumed_samples

            # User may pass in too-large
            if y is not None and len(y) == len(x):
                warn(
                    f"The first {n_consumed_samples} entries of y were ignored because "
                    "TimeDelay obesrvables were used."
                )
                y = y[n_consumed_samples:]
        else:
            n_consumed_samples = 0

        if y is None:
            if cast_as_real:
                return metric(
                    x[n_consumed_samples + 1 :].real,
                    self.predict(x[:-1]).real,
                    **metric_kws,
                )
            else:
                return metric(
                    x[n_consumed_samples + 1 :], self.predict(x[:-1]), **metric_kws
                )
        else:
            if cast_as_real:
                return metric(y.real, self.predict(x).real, **metric_kws)
            else:
                return metric(y, self.predict(x), **metric_kws)

    def get_feature_names(self, input_features=None):
        """
        Get the names of the individual features constituting the observables.

        Parameters
        ----------
        input_features: list of string, length n_input_features, \
                optional (default None)
            String names for input features, if available. By default,
            the names "x0", "x1", ... ,"xn_input_features" are used.

        Returns
        -------
        output_feature_names: list of string, length n_ouput_features
            Output feature names.
        """
        check_is_fitted(self, "n_input_features_")
        return self.observables.get_feature_names(input_features=input_features)

    def _step(self, x):
        """
        Map x one timestep forward in the space of observables.

        Parameters
        ----------
        x: numpy.ndarray, shape (n_samples, n_input_features)
            State vectors to be stepped forward.

        Returns
        -------
        X': numpy.ndarray, shape (n_samples, self.n_output_features_)
            Observables one timestep after x.
        """
        check_is_fitted(self, "n_output_features_")
        return self.model.predict(x)

    @property
    def koopman_matrix(self):
        """
        The Koopman matrix K satisfying g(X') = g(X) * K
        where g denotes the observables map and X' denotes x advanced one timestep.
        """
        check_is_fitted(self, "n_output_features_")
        return self.model.steps[-1][1].coef_



import pytest
from sklearn.utils.validation import check_is_fitted

from pykoopman import KoopmanContinuous
from pykoopman.differentiation import Derivative


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_random"), pytest.lazy_fixture("data_random_complex")],
)
def test_derivative_integration(data):
    x = data

    diff = Derivative(kind="finite_difference", k=1)
    model = KoopmanContinuous(differentiator=diff)

    model.fit(x)
    check_is_fitted(model)



"""
Linear observables
"""
from itertools import combinations
from itertools import combinations_with_replacement

from numpy import empty
from sklearn.utils.validation import check_is_fitted

from ..common import validate_input
from ._base import BaseObservables


class CustomObservables(BaseObservables):
    """
    Custom observable functions.

    To ensure invertibility of the set of observables, the identity map is
    automatically included along with user-specified observables.

    Parameters
    ----------
    observables: list of callable
        A list of functions mapping from state variables to observables.
        Univariate functions are applied to each state variable in turn and
        multivariable functions are applied to each combination of state
        variables.
        Note that the identity map is automatically prepended to this list.

    observable_names: list of callable, optional (default None)
        A list of functions mapping from names of state variables to names
        of observables. For example, the observable name
        :code:`lambda x: f"{x}^2"` would correspond to the function :math:`x^2`.
        If None, the names "f0(...)", "f1(...)", ... will be used.

    interaction_only: bool, optional (default True)
        Whether to omit self-interaction terms.
        If True, function evaluations of the form :math:`f(x,x)` and :math:`f(x,y,x)`
        will be omitted, but those of the form :math:`f(x,y)` and :math:`f(x,y,z)`
        will be included.
        If False, all combinations will be included.

    Attributes
    ----------
    n_input_features_ : int
        Number of input features.

    n_output_features_ : int
        Number of output features.
    """

    def __init__(self, observables, observable_names=None, interaction_only=True):
        super(CustomObservables, self).__init__()
        self.observables = [identity, *observables]
        if observable_names and (len(observables) != len(observable_names)):
            raise ValueError(
                "observables and observable_names must have the same length"
            )
        self.observable_names = observable_names
        self.interaction_only = interaction_only

    def fit(self, x, y=None):
        """
        Fit to measurement data.

        Determines the number of input and output features and creates
        default values for :code:`observable_names` if necessary.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Measurement data to be fit.

        y: None
            Dummy parameter retained for sklearn compatibility.

        Returns
        -------
        self: a fit :class:`pykoopman.observables.CustomObservables` instance
        """
        n_samples, n_features = validate_input(x).shape

        n_output_features = 0
        for f in self.observables:
            n_args = f.__code__.co_argcount
            n_output_features += len(
                list(self._combinations(n_features, n_args, self.interaction_only))
            )

        self.n_input_features_ = n_features
        self.n_output_features_ = n_output_features

        if self.observable_names is None:
            self.observable_names = list(
                map(
                    lambda i: (lambda *x: "f" + str(i) + "(" + ",".join(x) + ")"),
                    range(len(self.observables)),
                )
            )
        # First map is the identity
        self.observable_names.insert(0, identity_name)

        return self

    def transform(self, x):
        """
        Apply custom transformations to data, computing observables.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Measurement data to be transformed.

        Returns
        -------
        y: array-like, shape (n_samples, n_output_features)
            Transformed data (observables).
        """
        check_is_fitted(self, "n_input_features_")
        x = validate_input(x)

        n_samples, n_features = x.shape

        if n_features != self.n_input_features_:
            raise ValueError("x.shape[1] does not match n_input_features_")

        x_transformed = empty((n_samples, self.n_output_features_), dtype=x.dtype)
        observables_idx = 0
        for f in self.observables:
            for c in self._combinations(
                self.n_input_features_, f.__code__.co_argcount, self.interaction_only
            ):
                x_transformed[:, observables_idx] = f(*[x[:, j] for j in c])
                observables_idx += 1

        return x_transformed

    def inverse(self, y):
        """
        Invert the transformation.

        This function satisfies
        :code:`self.inverse(self.transform(x)) == x`

        Parameters
        ----------
        y: array-like, shape (n_samples, n_output_features)
            Data to which to apply the inverse.
            Must have the same number of features as the transformed data

        Returns
        -------
        x: array-like, shape (n_samples, n_input_features)
            Output of inverse map applied to y.
            In this case, x is identical to y.
        """
        y = validate_input(y)
        check_is_fitted(self, "n_input_features_")
        if y.shape[1] != self.n_output_features_:
            raise ValueError(
                "Wrong number of input features."
                f"Expected y.shape[1] = {self.n_out_features_}; "
                f"instead y.shape[1] = {y.shape[1]}."
            )
        # self.observables[0] is the identity, so original features are
        # just the first self.n_input_features_ columns
        return y[:, : self.n_input_features_]

    def get_feature_names(self, input_features=None):
        """
        Get the names of the output features.

        Parameters
        ----------
        input_features: list of string, length n_input_features, \
                optional (default None)
            String names for input features, if available. By default,
            the names "x0", "x1", ... ,"xn_input_features" are used.

        Returns
        -------
        output_feature_names: list of string, length n_ouput_features
            Output feature names.
        """
        check_is_fitted(self, "n_input_features_")
        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )

        feature_names = []
        for i, f in enumerate(self.observables):
            feature_names.extend(
                [
                    self.observable_names[i](*[input_features[j] for j in c])
                    for c in self._combinations(
                        self.n_input_features_,
                        f.__code__.co_argcount,
                        self.interaction_only,
                    )
                ]
            )

        return feature_names

    @staticmethod
    def _combinations(n_features, n_args, interaction_only):
        """Get the combinations of features to be passed to observable functions."""
        comb = combinations if interaction_only else combinations_with_replacement
        return comb(range(n_features), n_args)


def identity(x):
    """Identity map."""
    return x


def identity_name(x):
    """Name for identity map."""
    return str(x)



from pydmd import DMD
from pydmd import DMDBase
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from .common.base import validate_input
from .observables import Polynomial
from .regression import BaseRegressor
from .regression import DMDRegressor


class Koopman:
    """Primary Koopman class."""

    def __init__(self, observables=None, regressor=None, dt_default=1):
        if observables is None:
            observables = Polynomial(degree=1)
        if regressor is None:
            regressor = DMD()
        if not isinstance(dt_default, float) and not isinstance(dt_default, int):
            raise ValueError("dt_default must be a positive number")
        elif dt_default <= 0:
            raise ValueError("dt_default must be a positive number")
        else:
            self.dt_default = dt_default

        self.observables = observables
        self.regressor = regressor

    def fit(self, x, x_dot=None, dt=None, x_shift=None):
        if dt is None:
            dt = self.dt_default

        x = validate_input(x, dt)
        if x_dot is None:
            x_dot = x[1:]
            x = x[:-1]
        else:
            x_dot = validate_input(x_dot, dt)

        if isinstance(self.regressor, DMDBase):
            regressor = DMDRegressor(self.regressor)
        else:
            regressor = BaseRegressor(self.regressor)

        steps = [
            ("observables", self.observables),
            ("regressor", regressor),
        ]
        self.model = Pipeline(steps)

        # TODO: make this solve the correct problem
        self.model.fit(x, x_dot)

        self.n_input_features_ = self.model.steps[0][1].n_input_features_
        self.n_output_features_ = self.model.steps[0][1].n_output_features_

        return self

    def predict(self, x):
        check_is_fitted(self, "model")
        return self.observables.inverse(self._step(x))

    def simulate(self, x, n_steps=1):
        check_is_fitted(self, "model")
        # Could have an option to only return the end state and not all
        # intermediate states to save memory.
        output = [self.predict(x)]
        for k in range(n_steps - 1):
            output.append(self.predict(output[-1]))

    def _step(self, x):
        # TODO: rename this
        check_is_fitted(self, "model")
        return self.model.predict(x)

    @property
    def koopman_matrix(self):
        """
        Get the Koopman matrix K such that
        g(X') = g(X) * K
        """
        check_is_fitted(self, "model")
        return self.model.steps[-1][1].coef_



"""
Time-delay observables
"""
from numpy import arange
from numpy import empty
from sklearn.utils.validation import check_is_fitted

from ..common import validate_input
from ._base import BaseObservables


class TimeDelay(BaseObservables):
    r"""
    Time-delay observables. Observables formed by taking time-lagged
    measurements of state variables and interpreting them as new state
    variables.

    For example, the two state variables :math:`[x(t), y(t)]` could be
    supplemented with two time-delays each, yielding a new set of
    observables:

    .. math::
        [x(t), y(t), x(t-\Delta t), y(t-\Delta t),
        x(t-2\Delta t), y(t - 2\Delta t)]

    This example corresponds to taking :code:`delay =` :math:`\Delta t`
    and :code:`n_delays = 2`.

    Note that when transforming data the first
    :code:`delay * n_delays` rows/samples are dropped as there is
    insufficient time history to form time-delays for them.

    See the following references for more information.

        Brunton, Steven L., et al.
        "Chaos as an intermittently forced linear system."
        Nature communications 8.1 (2017): 1-9.

        Susuki, Yoshihiko, and Igor Mezić.
        "A prony approximation of Koopman mode decomposition."
        2015 54th IEEE Conference on Decision and Control (CDC). IEEE, 2015.

        Arbabi, Hassan, and Igor Mezic.
        "Ergodic theory, dynamic mode decomposition, and computation
        of spectral properties of the Koopman operator."
        SIAM Journal on Applied Dynamical Systems 16.4 (2017): 2096-2126.


    Parameters
    ----------
    delay: nonnegative integer, optional (default 1)
        The length of each delay.

    n_delays: nonnegative integer, optional (default 2)
        The number of delays to compute for each variable.

    Attributes
    ----------
    n_input_features_ : int
        Number of input features.

    n_output_features_ : int
        Number of output features.

    _n_consumed_samples : int
        Number of samples consumed when :code:`transform` is called,
        i.e. :code:`n_delays * delay`.
    """

    def __init__(self, delay=1, n_delays=2):
        super(TimeDelay, self).__init__()
        if delay < 0:
            raise ValueError("delay must be a nonnegative int")
        if n_delays < 0:
            raise ValueError("n_delays must be a nonnegative int")

        self.delay = int(delay)
        self.n_delays = int(n_delays)
        self._n_consumed_samples = self.delay * self.n_delays

    def fit(self, x, y=None):
        """
        Fit to measurement data.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Measurement data to be fit.

        y: None
            Dummy parameter retained for sklearn compatibility.

        Returns
        -------
        self: returns a fit :class:`pykoopman.observables.TimeDelay` instance
        """
        n_samples, n_features = validate_input(x).shape

        self.n_input_features_ = n_features
        self.n_output_features_ = n_features * (1 + self.n_delays)

        return self

    def transform(self, x):
        """
        Add time-delay features to the data, dropping the first
        :code:`delay - n_delays` samples.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Measurement data to be transformed.
            It is assumed that rows correspond to examples that are equi-spaced
            in time and are in sequential order.

        Returns
        -------
        y: array-like, shape (n_samples - delay * n_delays, n_output_features)
            Transformed data. Note that the number of output examples is,
            in general, different from the number input. In particular,
            :code:`n_samples - delay * n_delays`
        """
        check_is_fitted(self, "n_input_features_")
        x = validate_input(x)

        if x.shape[1] != self.n_input_features_:
            raise ValueError(
                "Wrong number of input features. "
                f"Expected x.shape[1] = {self.n_input_features_}; "
                f"instead x.shape[1] = {x.shape[1]}."
            )

        self._n_consumed_samples = self.delay * self.n_delays
        if len(x) < self._n_consumed_samples + 1:
            raise ValueError(
                "x has too few rows. To compute time-delay features with "
                f"delay = {self.delay} and n_delays = {self.n_delays} "
                f"x must have at least {self._n_consumed_samples + 1} rows."
            )

        y = empty(
            (x.shape[0] - self._n_consumed_samples, self.n_output_features_),
            dtype=x.dtype,
        )
        y[:, : self.n_input_features_] = x[self._n_consumed_samples :]

        for i in range(self._n_consumed_samples, x.shape[0]):
            y[i - self._n_consumed_samples, self.n_input_features_ :] = x[
                self._delay_inds(i), :
            ].flatten()

        return y

    def inverse(self, y):
        """
        Invert the transformation.

        This function satisfies
        :code:`self.inverse(self.transform(x)) == x[self._n_consumed_samples:]`

        Parameters
        ----------
        y: array-like, shape (n_samples, n_output_features)
            Data to which to apply the inverse.
            Must have the same number of features as the transformed data

        Returns
        -------
        x: array-like, shape (n_samples, n_input_features)
            Output of inverse map applied to y.
        """
        check_is_fitted(self, "n_input_features_")
        if y.shape[1] != self.n_output_features_:
            raise ValueError(
                "Wrong number of input features."
                f"Expected y.shape[1] = {self.n_out_features_}; "
                f"instead y.shape[1] = {y.shape[1]}."
            )

        # The first n_input_features_ columns correspond to the un-delayed
        # measurements
        return y[:, : self.n_input_features_]

    def get_feature_names(self, input_features=None):
        """
        Get the names of the output features.

        Parameters
        ----------
        input_features: list of string, length n_input_features,\
         optional (default None)
            String names for input features, if available. By default,
            the names "x0", "x1", ... ,"xn_input_features" are used.

        Returns
        -------
        output_feature_names: list of string, length n_ouput_features
            Output feature names.
        """
        check_is_fitted(self, "n_input_features_")
        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )

        output_features = [f"{xi}(t)" for xi in input_features]
        output_features.extend(
            [
                f"{xi}(t-{i * self.delay}dt)"
                for i in range(1, self.n_delays + 1)
                for xi in input_features
            ]
        )

        return output_features

    def _delay_inds(self, index):
        return index - self.delay * arange(1, self.n_delays + 1)

    @property
    def n_consumed_samples(self):
        """
        The number of samples that are consumed as "initial conditions" for
        other samples. I.e. the number of samples for which time delays cannot
        be computed.
        """
        return self._n_consumed_samples



"""
Polynomial observables
"""
from itertools import chain
from itertools import combinations
from itertools import combinations_with_replacement as combinations_w_r

import numpy as np
from scipy import sparse
from sklearn.preprocessing import PolynomialFeatures
from sklearn.preprocessing._csr_polynomial_expansion import _csr_polynomial_expansion
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import FLOAT_DTYPES

from ..common import check_array


class Polynomial(PolynomialFeatures):
    """Polynomial observables.

    This is essentially just :code:`sklearn.preprocessing.PolynomialFeatures`
    with support for complex numbers.

    Parameters
    ----------
    degree : int, optional (default 2)
        The degree of the polynomial features.
    interaction_only : boolean, optional (default False)
        If true, only interaction features are produced: features that are
        products of at most ``degree`` *distinct* input features (so not
        ``x[1] ** 2``, ``x[0] * x[2] ** 3``, etc.).
    include_bias : boolean, optional (default True)
        If True (default), then include a bias column, the feature in which
        all polynomial powers are zero (i.e. a column of ones - acts as an
        intercept term in a linear model).
    order : str in {'C', 'F'}, optional (default 'C')
        Order of output array in the dense case. 'F' order is faster to
        compute, but may slow down subsequent estimators.
    """

    def __init__(self, degree=2, interaction_only=False, include_bias=True, order="C"):
        if degree == 0:
            raise ValueError(
                "degree must be at least 1, otherwise inverse cannot be computed"
            )
        super(Polynomial, self).__init__(
            degree=degree,
            interaction_only=interaction_only,
            include_bias=include_bias,
            order=order,
        )

    def fit(self, x, y=None):
        """
        Compute number of output features.

        Parameters
        ----------
        x : np.ndarray, shape (n_samples, n_features)
            Measurement data.

        y : array-like, optional (default None)
            Dummy input.
        Returns
        -------
        self : fit :class:`pykoopman.observables.Polynomial` instance
        """
        return super(Polynomial, self).fit(x.real, y)

    def transform(self, x):
        """Transform data to polynomial features.

        Parameters
        ----------
        x : array-like or CSR/CSC sparse matrix, shape (n_samples, n_features)
            The data to transform, row by row.
            Prefer CSR over CSC for sparse input (for speed), but CSC is
            required if the degree is 4 or higher. If the degree is less than
            4 and the input format is CSC, it will be converted to CSR, have
            its polynomial features generated, then converted back to CSC.
            If the degree is 2 or 3, the method described in "Leveraging
            Sparsity to Speed Up Polynomial Feature Expansions of CSR Matrices
            Using K-Simplex Numbers" by Andrew Nystrom and John Hughes is
            used, which is much faster than the method used on CSC input. For
            this reason, a CSC input will be converted to CSR, and the output
            will be converted back to CSC prior to being returned, hence the
            preference of CSR.

        Returns
        -------
        xp : np.ndarray or CSR/CSC sparse matrix, shape (n_samples, n_output_features)
            The matrix of features, where n_output_features is the number of polynomial
            features generated from the combination of inputs.
        """
        check_is_fitted(self, "n_input_features_")

        x = check_array(x, order="F", dtype=FLOAT_DTYPES, accept_sparse=("csr", "csc"))

        n_samples, n_features = x.shape

        if n_features != self.n_input_features_:
            raise ValueError("x shape does not match training shape")

        if sparse.isspmatrix_csr(x):
            if self.degree > 3:
                return self.transform(x.tocsc()).tocsr()
            to_stack = []
            if self.include_bias:
                to_stack.append(np.ones(shape=(n_samples, 1), dtype=x.dtype))
            to_stack.append(x)
            for deg in range(2, self.degree + 1):
                xp_next = _csr_polynomial_expansion(
                    x.data,
                    x.indices,
                    x.indptr,
                    x.shape[1],
                    self.interaction_only,
                    deg,
                )
                if xp_next is None:
                    break
                to_stack.append(xp_next)
            xp = sparse.hstack(to_stack, format="csr")
        elif sparse.isspmatrix_csc(x) and self.degree < 4:
            return self.transform(x.tocsr()).tocsc()
        else:
            combinations = self._combinations(
                n_features,
                self.degree,
                self.interaction_only,
                self.include_bias,
            )
            if sparse.isspmatrix(x):
                columns = []
                for comb in combinations:
                    if comb:
                        out_col = 1
                        for col_idx in comb:
                            out_col = x[:, col_idx].multiply(out_col)
                        columns.append(out_col)
                    else:
                        bias = sparse.csc_matrix(np.ones((x.shape[0], 1)))
                        columns.append(bias)
                xp = sparse.hstack(columns, dtype=x.dtype).tocsc()
            else:
                xp = np.empty(
                    (n_samples, self.n_output_features_),
                    dtype=x.dtype,
                    order=self.order,
                )
                for i, comb in enumerate(combinations):
                    xp[:, i] = x[:, comb].prod(1)

        return xp

    def inverse(self, y):
        """
        Invert the transformation.

        This function satisfies
        :code:`self.inverse(self.transform(x)) == x`

        Parameters
        ----------
        y: numpy.ndarray, shape (n_samples, n_output_features)
            Data to which to apply the inverse.
            Must have the same number of features as the transformed data

        Returns
        -------
        x: numpy.ndarray, shape (n_samples, n_input_features)
            Output of inverse map applied to y.
        """
        check_is_fitted(self, "n_output_features_")
        if y.shape[1] != self.n_output_features_:
            raise ValueError(
                "y has the wrong number of features (columns)."
                f"Expected {self.n_output_features_}, received {y.shape[1]}"
            )
        if self.include_bias:
            return y[:, 1 : self.n_input_features_ + 1]
        else:
            return y[:, : self.n_input_features_]

    @staticmethod
    def _combinations(n_features, degree, interaction_only, include_bias):
        comb = combinations if interaction_only else combinations_w_r
        start = int(not include_bias)
        return chain.from_iterable(
            comb(range(n_features), i) for i in range(start, degree + 1)
        )



"""
Shared pytest fixtures for unit tests.

Put any datasets that are used by multiple unit test files here.
"""
import numpy as np
import pytest

from pykoopman.observables import CustomObservables


@pytest.fixture
def data_random():
    x = np.random.randn(50, 10)
    return x


@pytest.fixture
def data_random_complex():
    x = np.random.randn(50, 10) + 1j * np.random.randn(50, 10)
    return x


@pytest.fixture
def data_2D_superposition():
    t = np.linspace(0, 2 * np.pi, 200)
    x = np.linspace(-5, 5, 100)
    [x_grid, t_grid] = np.meshgrid(x, t)

    def sech(x):
        return 1 / np.cosh(x)

    f1 = sech(x_grid + 3) * np.exp(1j * 2.3 * t_grid)
    f2 = 2 * (sech(x_grid) * np.tanh(x_grid)) * np.exp(1j * 2.8 * t_grid)
    return f1 + f2


@pytest.fixture
def data_1D_cosine():
    t = np.linspace(0, 2 * np.pi, 200)
    x = np.cos(3 * t)
    return x


@pytest.fixture
def data_custom_observables():
    observables = [lambda x: x, lambda x: x ** 2, lambda x: 0 * x, lambda x, y: x * y]
    observable_names = [
        lambda s: str(s),
        lambda s: f"{s}^2",
        lambda s: str(0),
        lambda s, t: f"{s} {t}",
    ]

    return CustomObservables(observables, observable_names=observable_names)


@pytest.fixture
def data_realistic_custom_observables():
    observables = [lambda x: x ** 2, lambda x, y: x * y]
    observable_names = [
        lambda s: f"{s}^2",
        lambda s, t: f"{s} {t}",
    ]

    return CustomObservables(observables, observable_names=observable_names)



from numpy import arange
from pydmd import DMD
from pydmd import DMDBase
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from .common import drop_nan_rows
from .common import validate_input
from .differentiation import Derivative
from .observables import Identity
from .regression import BaseRegressor
from .regression import DMDRegressor


class KoopmanContinuous(BaseEstimator):
    """
    Continuous-time Koopman class.

    Parameters
    ----------
    observables: observables object, optional \
            (default :class:`pykoopman.observables.Identity`)
        Map(s) to apply to raw measurement data before estimating the
        Koopman operator.
        Must extend :class:`pykoopman.observables.BaseObservables`.
        The default option, :class`pykoopman.observables.Identity` leaves
        the input untouched.

    differentiator: callable, optional (default centered difference)
        Function used to compute numerical derivatives. The function must
        have the call signature :code:`differentiator(x, t)`, where ``x`` is
        a 2D numpy ndarray of shape ``(n_samples, n_features)`` and ``t`` is
        a 1D numpy ndarray of shape ``(n_samples,)``.

    regressor: regressor object, optional (default ``DMD``)
        The regressor used to learn the Koopman operator from the observables.
        ``regressor`` can either extend
        :class:`pykoopman.regression.BaseRegressor`, or the ``pydmd.DMDBase``
        class. In the latter case, the pydmd object must have both a ``fit``
        and a ``predict`` method.

    TODO
    """

    def __init__(
        self, observables=None, differentiator=None, regressor=None, dt_default=1
    ):
        if observables is None:
            observables = Identity()
        if differentiator is None:
            differentiator = Derivative(kind="finite_difference", k=1)
        if regressor is None:
            regressor = DMD()
        if not isinstance(dt_default, float) and not isinstance(dt_default, int):
            raise ValueError("dt_default must be a positive number")
        elif dt_default <= 0:
            raise ValueError("dt_default must be a positive number")
        else:
            self.dt_default = dt_default

        self.observables = observables
        self.differentiator = differentiator
        self.regressor = regressor

    def fit(self, x, x_dot=None, dt=None, t=None, x_shift=None):
        if dt is None:
            dt = self.dt_default

        if t is None:
            t = dt * arange(x.shape[0])

        # TODO: validate data
        x = validate_input(x)

        # TODO: this will probably need to change as we need to compute derivatives
        # after computing observables
        if x_dot is None:
            x_dot = self.differentiator(x, t)

        # Some differentiation methods generate NaN entries at endpoints
        x_dot, x = drop_nan_rows(x_dot, x)

        if isinstance(self.regressor, DMDBase):
            regressor = DMDRegressor(self.regressor)
        else:
            regressor = BaseRegressor(self.regressor)

        steps = [
            ("observables", self.observables),
            ("regressor", regressor),
        ]
        self.model = Pipeline(steps)

        # TODO: make this solves the correct problem
        self.model.fit(x, x_dot)

        self.n_input_features_ = self.model.steps[0][1].n_input_features_
        self.n_output_features_ = self.model.steps[0][1].n_output_features_

        return self

    def predict(self, x):
        check_is_fitted(self, "model")
        return self.observables.inverse(self._step(x))

    def simulate(self, x, n_steps=1):
        check_is_fitted(self, "model")
        # Could have an option to only return the end state and not all
        # intermediate states to save memory.
        output = [self.predict(x)]
        for k in range(n_steps - 1):
            output.append(self.predict(output[-1]))

    def _step(self, x):
        # TODO: rename this
        check_is_fitted(self, "model")
        return self.model.predict(x)

    @property
    def koopman_matrix(self):
        """
        Get the Koopman matrix K such that
        g(X') = g(X) * K
        """
        check_is_fitted(self, "model")
        return self.model.steps[-1][1].coef_



"""Tests for (discrete) Koopman objects."""
import pytest
from numpy.testing import assert_allclose
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from pykoopman import Koopman
from pykoopman.observables import Identity
from pykoopman.observables import Polynomial
from pykoopman.observables import TimeDelay


def test_fit(data_random):
    x = data_random
    model = Koopman().fit(x)
    check_is_fitted(model)


def test_predict_shape(data_random):
    x = data_random

    model = Koopman().fit(x)
    assert x.shape == model.predict(x).shape


def test_simulate_accuracy(data_2D_superposition):
    x = data_2D_superposition

    model = Koopman().fit(x)

    n_steps = 10
    x_pred = model.simulate(x[0], n_steps=n_steps)
    assert_allclose(x[1 : n_steps + 1], x_pred)


def test_koopman_matrix_shape(data_random):
    x = data_random
    model = Koopman().fit(x)
    assert model.koopman_matrix.shape[0] == model.n_output_features_


def test_if_fitted(data_random):
    x = data_random
    model = Koopman()

    with pytest.raises(NotFittedError):
        model.predict(x)
    with pytest.raises(NotFittedError):
        model.simulate(x)
    with pytest.raises(NotFittedError):
        model.koopman_matrix
    with pytest.raises(NotFittedError):
        model._step(x)
    with pytest.raises(NotFittedError):
        model.score(x)


def test_score_without_target(data_2D_superposition):
    x = data_2D_superposition
    model = Koopman().fit(x)

    # Test without a target
    assert model.score(x) > 0.8


def test_score_with_target(data_2D_superposition):
    x = data_2D_superposition
    model = Koopman().fit(x)

    # Test with a target
    assert model.score(x[::2], y=x[1::2]) > 0.8


def test_score_complex_data(data_random_complex):
    x = data_random_complex
    model = Koopman().fit(x)

    with pytest.raises(ValueError):
        model.score(x, cast_as_real=False)


@pytest.mark.parametrize(
    "observables",
    [
        Identity(),
        Polynomial(),
        TimeDelay(),
        pytest.lazy_fixture("data_custom_observables"),
    ],
)
def test_observables_integration(data_random, observables):
    x = data_random
    model = Koopman(observables=observables).fit(x)
    check_is_fitted(model)

    y = model.predict(x)
    assert y.shape[1] == x.shape[1]


@pytest.mark.parametrize(
    "observables",
    [
        Identity(),
        Polynomial(),
        TimeDelay(),
        pytest.lazy_fixture("data_custom_observables"),
    ],
)
def test_observables_integration_accuracy(data_1D_cosine, observables):
    x = data_1D_cosine
    model = Koopman(observables=observables, quiet=True).fit(x)

    assert model.score(x) > 0.95


def test_simulate_with_time_delay(data_2D_superposition):
    x = data_2D_superposition

    observables = TimeDelay()
    model = Koopman(observables=observables)
    model.fit(x)

    n_steps = 10
    n_consumed_samples = observables.n_consumed_samples
    x_pred = model.simulate(x[: n_consumed_samples + 1], n_steps=n_steps)
    assert_allclose(
        x[n_consumed_samples + 1 : n_consumed_samples + n_steps + 1], x_pred
    )
